# THIS FILE IS GENERATED FROM ENSTALLER SETUP.PY
version = '4.8.7'
full_version = '4.8.7'
git_revision = '4ba6fba3034936c6cfcacd878a7cff3c81dda174'
is_released = True

if not is_released:
    version = full_version
